#ifndef FrameCPP_VERSION_4_FrDetector_HH
#define FrameCPP_VERSION_4_FrDetector_HH

#include <string>
#include <vector>

#include "general/types.hh"

#include "framecpp/FrameCPP.hh"

#include "framecpp/Common/SearchContainer.hh"
#include "framecpp/Common/TOCInfo.hh"
#include "framecpp/Common/FrDetector.hh"

#include "framecpp/Version4/GPSTime.hh"
#include "framecpp/Version4/STRING.hh"

#include "framecpp/Version4/FrTable.hh"
#include "framecpp/Version4/FrVect.hh"

#include "framecpp/Version3/FrDetector.hh"

namespace FrameCPP
{
  namespace Version_4
  {
    class FrDetector
      : public Common::FrDetector,
	public Common::TOCInfo
    {
    public:
      typedef Common::SearchContainer< FrVect, &FrVect::GetName >
      more_type;
      typedef more_type::iterator aux_iterator;
      typedef more_type::const_iterator const_aux_iterator;

      typedef Common::SearchContainer< FrTable, &FrTable::GetNameSlow >
      more_table_type;
      typedef more_table_type::iterator table_iterator;
      typedef more_table_type::const_iterator const_table_iterator;

      //: Default constructor
      FrDetector( );

      FrDetector( const std::string& Name,
		  const INT_2S LongitudeD,
		  const INT_2S LongitudeM,
		  const REAL_4 LongitudeS,
		  const INT_2S LatitudeD,
		  const INT_2S LatitudeM,
		  const REAL_4 LatitudeS,
		  const REAL_4 Elevation,
		  const REAL_4 ArmXazimuth,
		  const REAL_4 ArmYazimuth );

      FrDetector( const FrDetector& detector );

      FrDetector( Previous::FrDetector& Source,
		  Common::IStream* Stream );

      static const char* const StructName( );

      static const Common::Description* StructDescription( );

      //: Number of bytes needed to write this structure
      cmn_streamsize_type Bytes( const Common::StreamBase& Stream) const;

      virtual FrDetector* Create( Common::IStream& Stream ) const;

      virtual const char* const ObjectStructName( ) const;

      virtual void Write( Common::OStream& Stream ) const;

      const std::string& GetName( ) const;
      INT_2S GetLongitudeD( ) const;
      INT_2S GetLongitudeM( ) const;
      REAL_4 GetLongitudeS( ) const;
      INT_2S GetLatitudeD( ) const;
      INT_2S GetLatitudeM( ) const;
      REAL_4 GetLatitudeS( ) const;
      REAL_4 GetElevation( ) const;
      REAL_4 GetArmXazimuth( ) const;
      REAL_4 GetArmYazimuth( ) const;

      const more_type& RefMore( ) const;
      more_type& RefMore( );
      const more_table_type& RefMoreTable( ) const;
      more_table_type& RefMoreTable( );
   
      //: Merge with another FrDetector
      FrDetector& Merge( const FrDetector& RHS );

      static std::auto_ptr< Object >
      Promote( INT_2U Source,
	       Object* Obj, Common::IStream* Stream );

      /// \brief comparison operator
      bool operator==( const FrDetector& RHS ) const;

      /// \brief comparison operator
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      virtual void TOCQuery( int InfoClass, ... ) const;

    protected:
      /// \brief Down grade an object
      virtual std::auto_ptr< Object >
      demote( INT_2U Target, Common::IStream* Stream ) const;

      /// \brief Upgrade an object
      virtual std::auto_ptr< Object >
      promote( INT_2U Target, Object* Obj, Common::IStream* Stream ) const;

    private:
      using Common::FrameSpec::Object::Create;

      struct fr_detector_data_type {
	bool operator==( const fr_detector_data_type& RHS ) const;
	STRING		name;
	INT_2S		longitudeD;
	INT_2S		longitudeM;
	REAL_4		longitudeS;
	INT_2S		latitudeD;
	INT_2S		latitudeM;
	REAL_4		latitudeS;
	REAL_4		elevation;
	REAL_4		armXazimuth;
	REAL_4		armYazimuth;
	more_type	more;
	more_table_type	moreTable;	
      };

      fr_detector_data_type	m_data;

      FrDetector( Common::IStream& Stream );
    };

    inline const char* const FrDetector::
    StructName( )
    {
      static const char* const name = "FrDetector";
      return name;
    }

    inline INT_2S FrDetector::
    GetLongitudeD( ) const
    {
      return m_data.longitudeD;
    }

    inline INT_2S FrDetector::
    GetLongitudeM( ) const
    {
      return m_data.longitudeD;
    }

    inline REAL_4 FrDetector::
    GetLongitudeS( ) const
    {
      return m_data.longitudeS;
    }

    inline INT_2S FrDetector::
    GetLatitudeD( ) const
    {
      return m_data.latitudeD;
    }

    inline INT_2S FrDetector::
    GetLatitudeM( ) const
    {
      return m_data.latitudeM;
    }

    inline REAL_4 FrDetector::
    GetLatitudeS( ) const
    {
      return m_data.latitudeS;
    }

    inline REAL_4 FrDetector::
    GetElevation( ) const
    {
      return m_data.elevation;
    }

    inline REAL_4 FrDetector::
    GetArmXazimuth( ) const
    {
      return m_data.armXazimuth;
    }

    inline REAL_4 FrDetector::
    GetArmYazimuth( ) const
    {
      return m_data.armYazimuth;
    }

    inline const FrDetector::more_type& FrDetector::
    RefMore( ) const
    {
      return m_data.more;
    }

    inline FrDetector::more_type& FrDetector::
    RefMore( )
    {
      return m_data.more;
    }

    inline const FrDetector::more_table_type& FrDetector::
    RefMoreTable( ) const
    {
      return m_data.moreTable;
    }

    inline FrDetector::more_table_type& FrDetector::
    RefMoreTable( )
    {
      return m_data.moreTable;
    }
   
    inline std::auto_ptr< Common::FrameSpec::Object > FrDetector::
    Promote( INT_2U Source,
	     Common::FrameSpec::Object* Obj, Common::IStream* Stream )
    {
      return Object::PromoteObject< Previous::FrDetector, FrDetector >
	( DATA_FORMAT_VERSION, Source, Obj, Stream );
    }

    inline bool FrDetector::
    operator==( const FrDetector& RHS ) const
    {
      return ( m_data == RHS.m_data );
    }
  } // namespace - Version_4
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_4_FrDetector_HH */
