#include "framecpp/config.h"

#include "framecpp/Common/Description.hh"
#include "framecpp/Common/IOStream.hh"

#include "framecpp/Version6/FrameSpec.hh"
#include "framecpp/Version6/FrTable.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"

using namespace FrameCPP::Version_6;
using FrameCPP::Common::Description;
using FrameCPP::Common::FrameSpec;
using FrameCPP::Common::IStream;
using FrameCPP::Common::OStream;

static std::auto_ptr< Description > init_description( );

static std::auto_ptr< Description > s_description = init_description( );

//=======================================================================
//=======================================================================

static const INT_2U MAX_REF = 1;

FrTableData::
FrTableData( )
{
}

FrTableData::
FrTableData( const FrTableData& Source )
  : m_data( Source.m_data )
{
}

//=======================================================================
//=======================================================================
FrTable::
FrTable( )
  : Object( CLASS_ID, s_description.get( ) )
{
}

FrTable::
FrTable( const FrTable& Source )
  : Object( CLASS_ID, s_description.get( ) ),
    FrTableData( Source )
{
  RefColumn( ) = Source.RefColumn( );
}

FrTable::
FrTable( const FrTableData& Source )
  : Object( CLASS_ID, s_description.get( ) ),
    FrTableData( Source )
{
}

FrTable::
FrTable( const std::string& name, INT_4U nrows )
  : Object( CLASS_ID, s_description.get( ) )
{
  m_data.name = name;
  m_data.nRow = nrows;
}

FrTable::
FrTable( IStream& Stream )
  : Object( CLASS_ID, s_description.get( ) )
{
  INT_2U nColumn;

  Stream >> m_data.name
	 >> m_data.comment
	 >> nColumn
	 >> m_data.nRow
	 >> m_data.columnName
	 >> m_column
    ;
  Stream.Next( this );
}

FrTable::
FrTable( Previous::FrTable& Source, IStream* Stream )
  : Object( CLASS_ID, s_description.get( ) ),
    FrTableData( Source )
{
  if ( Stream )
  {
    //-------------------------------------------------------------------
    // Fix references
    //-------------------------------------------------------------------
    Stream->ReplaceRef( Source.RefColumn( ),
			&(RefColumn( )), MAX_REF );

  }
}

FrTable::
~FrTable( )
{
}

FrameCPP::cmn_streamsize_type  FrTable::
Bytes( const Common::StreamBase& Stream ) const
{
  cmn_streamsize_type retval
    = m_data.Bytes( Stream )
    + Stream.PtrStructBytes( )	// next
    ;
  return retval;
}

FrTable* FrTable::
Create( IStream& Stream ) const
{
  return new FrTable( Stream );
}

FrTable& FrTable::
Merge( const FrTable& RHS )
{
  throw
    Unimplemented( "FrTable& FrTable::Merge( const FrTable& RHS )",
		   DATA_FORMAT_VERSION, __FILE__, __LINE__ );
  return *this;
}

const char* const FrTable::
ObjectStructName( ) const
{
  return StructName( );
}

void FrTable::
Write( OStream& Stream ) const
{
  INT_2U nColumn( m_data.columnName.size( ) );

  Stream << m_data.name
	 << m_data.comment
	 << nColumn
	 << m_data.nRow
    ;
  for ( std::vector< STRING >::const_iterator
	  cur = m_data.columnName.begin( ),
	  last = m_data.columnName.end( );
	cur != last;
	++cur )
  {
    Stream << *cur;
  }
  Stream << m_column
    ;
  WriteNext( Stream );
}

const std::string& FrTable::
GetNameSlow( ) const
{
  return GetName( );
}

bool FrTable::
operator==( const Common::FrameSpec::Object& Obj ) const
{
  return compare( *this, Obj );
}

FrTable& FrTable::
operator+=( const FrTable& RHS )
{
  throw Unimplemented( "FrTable& FrTable::operator+=( const FrTable& RHS )",
		       DATA_FORMAT_VERSION, __FILE__, __LINE__ );
}

FrameCPP::cmn_streamsize_type FrTable::fr_table_data_type::
Bytes( const Common::StreamBase& Stream ) const
{
  cmn_streamsize_type retval
    = name.Bytes( )
    + comment.Bytes( )
    + sizeof( INT_2U ) // nColumn
    + sizeof( nRow )
    + Stream.PtrStructBytes( ) 	// column
    ;

    //:TODO: columnName
  for ( std::vector< STRING >::const_iterator
	  cur = columnName.begin( ),
	  last = columnName.end( );	cur != last;
	++cur )
  {
    retval += cur->Bytes( );
  }
  return retval;
}


std::auto_ptr< FrameSpec::Object > FrTable::
demote( INT_2U Target, IStream* Stream ) const
{
  if ( Target >= DATA_FORMAT_VERSION )
  {
    return std::auto_ptr< FrameSpec::Object >
      ( const_cast< FrTable* >( this ) );
  }
  try
  {
    //-------------------------------------------------------------------
    // Create a newly demoted instance.
    //-------------------------------------------------------------------
    std::auto_ptr< Previous::FrTable >
      prev( new Previous::FrTable( *this ) );

    //-------------------------------------------------------------------
    // Modify references
    //-------------------------------------------------------------------
    if ( Stream )
    {
#define SRR__( R ) \
      Stream->ReplaceRef( prev->R( ), &(this->R( )), MAX_REF )

      SRR__(RefColumn);
#undef SRR__
    }
    //-------------------------------------------------------------------
    // Return demoted object
    //-------------------------------------------------------------------
    std::auto_ptr< FrameSpec::Object > retval( prev.get( ) );
    if ( retval.get( ) )
    {
      prev.release( );
    }
    return retval;
  }
  catch( ... )
  {
  }

  throw
    Unimplemented( "Object* FrTable::Demote( Object* Obj ) const",
		   DATA_FORMAT_VERSION, __FILE__, __LINE__ );
}

std::auto_ptr< FrameSpec::Object > FrTable::
promote( INT_2U Source, Object* Obj, IStream* Stream ) const
{
  return PromoteObject< Previous::FrTable, FrTable >
    ( DATA_FORMAT_VERSION, Source, Obj, Stream );
}

static std::auto_ptr< Description >
init_description( )
{
  std::auto_ptr< Description >	retval( new Description );

  (*retval)( FrSH( FrTable::StructName( ), FrTable::CLASS_ID,
		   "Table  Data Structure" ) );

  (*retval)( FrSE( "name", "STRING",
		   "") );
  (*retval)( FrSE( "comment", "STRING",
		   "") );
  (*retval)( FrSE( "nColumn", "INT_2U",
		   "") );
  (*retval)( FrSE( "nRow", "INT_4U",
		   "") );
  (*retval)( FrSE( "columnName", "STRING[nColumn]",
		   "") );
  (*retval)( FrSE( "column", PTR_STRUCT::Desc( FrVect::StructName( ) ),
		   "" ) );
  (*retval)( FrSE( "next", PTR_STRUCT::Desc( FrTable::StructName( ) ),
		   "" ) );

  return retval;
}
